#!/usr/bin/perl -w

# firewall module for dulog - gets ipchains and iptables logs.
# v0.2 9/20/01 Michael D. Stenner <mstenner@phy.duke.edu>
# made for dulog 0.2.x
# Added to CVS by icon
# $Id: firewall.mod.pl,v 1.1 2002/01/16 14:46:38 icon Exp $
#

# NOTE: because of the flexibility allowed by iptables logging, you may
# need to set the FIREWALL_IPTABLES_PATTERN option.  Do this by exporting
# a variable of that name from the dulog config file.  It should be a
# perl regular expression.

use Socket;
use strict;

##############################################################################
# read options from environment variables (set by dulog)
# all of these are environment variables that can be exported from the
# dulog config file.

#                         option name        default

# $ip?_pattern is a regex that marks a hit - first captured string is
# considered the "action". For example, if you use "iptables: DROP ",
# then you want the pattern 'iptables: (\S+)'
my $ipt_pattern = option('FIREWALL_IPTABLES_PATTERN', 'iptables: (\S+)');
my $ipc_pattern = option('FIREWALL_IPCHAINS_PATTERN', 'Packet log: \S+ (\S+)');
my $hr_format   = option('FIREWALL_HOST_PATTERN',     '%12s %8s %-30s');
my $or_format   = option('FIREWALL_OFFENDER_PATTERN', '%s -> %s');
my $ldebug      = option('FIREWALL_DEBUG',   undef); # firewall.mod debug

### general dulog options
my $debug       = option('DEBUG',            undef); # dulog in debug mode
my $quiet       = option("QUIET",            undef); # dulog in quiet mode
my $input       = option("LOGCAT",           undef); # file to read
my $output      = option("LOGREPORT",        undef); # file to write report to
my $weed        = option("WEED",             undef); # remove processed logs
my $confdir     = option("CONFDIR",          "../etc"); # where to find config
my $trojan_file = "$confdir/trojans.list"; # list of trojans and other nasties

##############################################################################
# process the options

# open the input file, $LOGCAT or read from stdin if LOGCAT isn't defined.
if (defined($input)) {
    open(INPUT, $input) or mlog(0, "cannot open input file $input\n");
    $weed = defined($weed) ? 1 : 0;
} else {
    *INPUT = *STDIN;
    $weed  = 0
}

# open output file, $LOGREPORT or write to STDOUT if LOGREPORT isn't defined.
if (defined($output)) {
    open(OUTPUT, ">$output") or mlog(0, "cannot open output file $output\n");
} else {
    *OUTPUT = *STDOUT;
}

# we'll write unprocessed logs to a replacement $LOGCAT if weeding is on.
if ($weed) {
    open(NEWIN, ">$input.new") or mlog(0, "cannot open tmp file $input.new\n");
}

# set up logging level.  NOTE: logs at level 0 (serious problems) are reported
# even when "QUIET" is on!
my $loglevel = (defined($ldebug) or defined($debug)) ? 2 : 1;
$loglevel = 0 if defined($quiet);
*LOGOUT = *STDOUT; # this is where the logging goes

##############################################################################
# do the magic (main)

my %hostrep; my %offrep; my %known_hosts;

# load the trojans list
my %trojans; loadtrojans();

mlog(1, "beginning to read input");
my $line;
while (defined($line = <INPUT>)) {
    my ($action, $host, $dport, $shost) = ('', undef, undef, undef);

    #####################################################
    # ----- TO ADD ANOTHER TYPE OF RECOGNIZED LOG ----- #
    # add a block to the following if-elsif sequence that
    #  a) recognizes the desired log out of _all_ log types
    #  b) optionally sets $action
    #  c) sets $host, $dport, $shost
    if ($line =~ /$ipt_pattern/) { ##### IPTABLES #####
       	$action = defined($1) ? $1 : '';
	($host, $dport, $shost) = iptables($line);
    } elsif ($line =~ /$ipc_pattern/) { ##### IPCHAINS #####
	$action = defined($1) ? $1 : '';
	($host, $dport, $shost) = ipchains($line);
    }

    if ($host and $shost and $dport) {
	# following variables must be set:
	#    $host    = local host
	#    $shost   = sending (or source) host
	#    $dport   = destination port

	# syslog-ng can report hosts in a more complicated way :)
	if ($host =~ m{[@/](\S+)}) { $host = $1; }

	# host report  --  report by scanned host
	my $hr = sprintf($hr_format, $host, $action, $dport);
	# step counter
	$hostrep{$hr} = exists($hostrep{$hr}) ? $hostrep{$hr} + 1 : 1;

	# offender report  --  report by offending scanner
	my $or = sprintf($or_format, $shost, $dport);
	# step counter
	$offrep{$or} = exists($offrep{$or}) ? $offrep{$or} + 1 : 1;
    } else {
	# not a recognized packet log
	print(NEWIN $line) if $weed; # dump to NEWIN only if we're weeding
    }
}

if (scalar keys(%hostrep)) {
    my $key;
    mlog(1, "writing output");
    print OUTPUT "::Scanned hosts::\n";
    foreach $key (sort(keys(%hostrep))) {
	printf(OUTPUT "%6d %s\n", $hostrep{$key}, $key);
    }
    print OUTPUT "\n::Offenders::\n";
    foreach $key (sort(keys(%offrep))) {
	printf(OUTPUT "%6d %s\n", $offrep{$key}, $key);
    }
} else {
    mlog(1, "no packets logged - nothing to report");
}
    
if ($weed) { # if we're weeding, replace the log file
    mlog(1, "replacing input file $input with $input.new");
    rename("$input.new", $input)
	or mlog(0, "cannot move file $input.new to $input");
}

###################################################################
# subroutines

sub ipchains {
    my $line = shift;

    #         ---- time -----    host          proto src              port    
    $line =~ /^\S+\s+\S+\s+\S+\s+(\S+) .*PROTO=(\d+) (.*):\d+ [\d\.]+:(\d+).*$/
	or return();
    my $host    = $1;
    my ($proto) = getprotobynumber($2)    or return();
    my $shost   = hostname($3)            or return();
    my $dport   = servicename($4, $proto) or return();

    return($host, $dport, $shost);
}

sub iptables {
    my $line = shift;

    $line =~ /^\S+\s+\S+\s+\S+\s+(\S+) .*?(IN=.*)$/
	or return();
    my $host = $1;
    my $data = parse_tables($2) or return();
	
    return($host, $data->{'DSTPORT'}, $data->{'SRCHOST'});
}

sub parse_tables {
    # parse an iptables log
    my $string = shift; # this is the log itself
    my %data; # will dump data into this

    my @array = split(/\s+/, $string); # split on whitespace
    foreach my $item (@array) {
	if ($item =~ /^(\w+)(=?)(.*)/) { # example: "IN=eth0"
	    $data{$1} = $3;
	}
    }

    my $proto = lc($data{'PROTO'}); 
    my $port = $data{'DPT'};
    
    $data{'SRCHOST'} = hostname($data{'SRC'});
    $data{'DSTPORT'} = servicename($port, $proto);

    return \%data;
}

sub mlog {
    # write the log message if it is at or below the current logging level

    # level 0 -- serious problem, should always be reported
    #            (logs are probably not being reported correctly)
    # level 1 -- notification of important events
    # level 2 -- debugging output, useful only when coding this module

    my $level = shift;
    print(LOGOUT "DULOG(firewall.mod): ", @_, "\n") if ($level <= $loglevel);
}

sub loadtrojans {
    # read the trojans list from the file
    mlog(1, "loading trojans list");
    open(TROJANS, $trojan_file) or
	mlog(0, "cannot read trojan file $trojan_file");

    my $line;
    while (defined($line = <TROJANS>)) {
	chomp $line;
	my ($proto, $port, $description) = split(/:/, $line);
	$trojans{"$proto:$port"} = $description;
    }
}

sub option {
    # process options passed in by environment variable
    my $op      = shift;
    my $default = shift;
    return(exists($ENV{$op}) ? $ENV{$op} : $default);
}

sub hostname {
    my $host = shift;

    # 123.234.63.23 -> blah.foo.bar.com
    # (or remain the ip if it doesn't resolve)

    if (exists($known_hosts{$host})) {
	return($known_hosts{$host});
    } else {
	my @host_a = gethostbyaddr(pack('CCCC', split(/\./, $host)), AF_INET);
	my $hostname = defined($host_a[0]) ? $host_a[0] : $host;
	$known_hosts{$host} = $hostname;
	return($hostname);
    }
}

sub servicename {
    my ($port, $proto) = @_;
    my $service;
    
    my @serv_a = getservbyport($port, $proto); # set service name
    if (scalar(@serv_a)) { # did we find anything
	$service = $serv_a[0] . " ($port/$proto)"; # yes? use that.
    } elsif (exists($trojans{$proto . ":" . $port})) { # no? trojan?
	# yes? use that
	$service = $trojans{$proto . ":" . $port} . " ($port/$proto)";
    } else { # still no?  Fine...
	$service .= $port . "/$proto"; # just report the port number
    }

    return($service);
}
